# -*- coding: utf-8 -*-
from __future__ import absolute_import

import copy
import pickle
import sys

from django.core.management import call_command
from django.test import TestCase
from django.utils.encoding import force_text

from konst import Constant, ConstantGroup, Constants, json
from konst.extras.drf.fields import ConstantChoiceField as DRFConstantChoiceField
from konst.tests.models import Apple
from rest_framework import serializers

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class ConstantTestCase(TestCase):
    def test_repr(self):
        self.assertEqual(repr(Apple.purposes.cooking), u"cooking (0)")
        self.assertEqual(repr(Apple.colours.red), u"red (FF0000)")

    def test_str(self):
        self.assertEqual(str(Apple.purposes.cooking), u"0")
        self.assertEqual(str(Apple.colours.red), u"FF0000")

    def test_comparison_with_int(self):
        self.assertEqual(Apple.purposes.cooking, 0)
        self.assertNotEqual(Apple.purposes.cooking, 1)

    def test_comparison_with_char(self):
        self.assertEqual(Apple.colours.red, u"FF0000")
        self.assertNotEqual(Apple.colours.red, u"FF0001")

    def test_comparison_with_constant(self):
        self.assertEqual(Apple.purposes.cooking, Apple.purposes.cooking)
        self.assertNotEqual(Apple.purposes.cooking, Apple.purposes.eating)
        self.assertEqual(Apple.colours.red, Apple.colours.red)
        self.assertNotEqual(Apple.colours.red, Apple.colours.green)

    def test_constant_in_group(self):
        self.assertEqual(
            Apple.purposes.culinary,
            set(
                [Apple.purposes.eating, Apple.purposes.cooking, Apple.purposes.juicing]
            ),
        )

    def test_two_groups_defined(self):
        pets = Constants(
            Constant(husky="husky"),
            Constant(poodle="poodle"),
            Constant(burmese="burmese"),
            Constant(ginger="ginger"),
            ConstantGroup("cat", ("burmese", "ginger")),
            ConstantGroup("dog", ("husky", "poodle")),
        )
        self.assertTrue(pets.husky.dog)
        self.assertFalse(pets.husky.cat)
        self.assertTrue(pets.poodle.dog)
        self.assertFalse(pets.poodle.cat)
        self.assertTrue(pets.burmese.cat)
        self.assertFalse(pets.burmese.dog)
        self.assertTrue(pets.ginger.cat)
        self.assertFalse(pets.ginger.dog)
        self.assertEqual(pets.dog, set([pets.husky, pets.poodle]))
        self.assertEqual(pets.cat, set([pets.burmese, pets.ginger]))


class ConstantChoiceFieldTestCase(TestCase):

    fixtures = ["test_apples"]

    def setUp(self):
        self.instance = Apple.objects.first()

    def test_create_and_save(self):
        gala = Apple.objects.create(
            name=u"Gala", colour=Apple.colours.red, purpose=Apple.purposes.eating
        )
        self.assertTrue(isinstance(gala.colour, Constant))
        self.assertTrue(isinstance(gala.purpose, Constant))
        self.assertEqual(gala.colour, Apple.colours.red)
        self.assertTrue(gala.colour.red)

    def test_from_db(self):
        instance = Apple.objects.all().first()
        self.assertTrue(isinstance(instance.colour, Constant))
        self.assertTrue(isinstance(instance.purpose, Constant))
        self.assertTrue(instance.colour.green)
        self.assertTrue(instance.purpose.cooking)

    def test_widget_force_text(self):
        self.assertEqual(
            force_text(self.instance.colour), u"{}".format(self.instance.colour.v)
        )
        self.assertEqual(
            force_text(self.instance.purpose), u"{}".format(self.instance.purpose.v)
        )

    def test_getattr_equality_check(self):
        self.assertTrue(self.instance.colour.green)
        self.assertFalse(self.instance.colour.red)
        self.assertTrue(self.instance.purpose.cooking)
        self.assertFalse(self.instance.purpose.eating)

    def test_getattr_equality_check_missing_attr(self):
        with self.assertRaises(AttributeError):
            self.instance.colour.donkey
        with self.assertRaises(AttributeError):
            self.instance.purposes.donkey

    def test_loaddata(self):
        # if the grannysmith exists we"re good
        # as the tests have already loaded the fixture
        instance = Apple.objects.all().first()
        self.assertEqual(instance.name, u"Granny Smith")
        self.assertTrue(instance.colour.green)
        self.assertTrue(instance.purpose.cooking)

    def test_dumpdata(self):
        """Dump our users, load as JSON and inspect."""
        sysout = sys.stdout
        sys.stdout = StringIO()
        call_command("dumpdata", "tests.apple")
        dumped_data = json.loads(sys.stdout.getvalue())
        sys.stdout = sysout
        self.assertEqual(
            dumped_data[0]["fields"]["colour"], u"{}".format(Apple.colours.green)
        )

    def test_copy(self):
        # https://nedbatchelder.com/blog/201010/surprising_getattr_recursion.html
        def f(*args, **kw):
            pass

        sys.settrace(f)
        self.assertEqual(sys.gettrace(), f)
        colour = Apple.colours.red
        self.assertTrue(colour.red)
        self.assertFalse(colour.green)
        colour_2 = copy.copy(colour)
        self.assertNotEqual(sys.gettrace(), None)
        self.assertTrue(colour_2.red)
        self.assertFalse(colour_2.green)

    def test_pickle(self):
        # very similar to the test_copy above
        c = pickle.dumps(Apple.colours.red)
        try:
            colour = pickle.loads(c)
        except RuntimeError as r:
            self.fail("Loading pickled constant raised RuntimeError: {}".format(r))
        else:
            self.assertTrue(colour.red)
            self.assertFalse(colour.green)

    def test_group_in_filter(self):
        self.assertEqual(
            Apple.objects.filter(purpose__in=Apple.purposes.culinary).count(), 1
        )
        instance = Apple.objects.filter(purpose__in=Apple.purposes.culinary).first()
        self.assertEqual(self.instance, instance)
        self.assertTrue(instance.purpose.culinary)


class ConstantJSONSerializerTestCase(TestCase):

    # Can"t make it serializable by default .. use our custom encoder
    def test_json_serializable(self):
        data = {
            # "colour": Apple.colours.red,
            "purpose": Apple.purposes.cooking
        }
        self.assertEqual(json.loads(json.dumps(data)), data)


class AppleSerializer(serializers.ModelSerializer):

    purpose = DRFConstantChoiceField(Apple.purposes)
    colour = DRFConstantChoiceField(Apple.colours)

    class Meta:
        model = Apple
        fields = ("name", "purpose", "colour")


class DRFConstantChoiceFieldTestCase(TestCase):

    fixtures = ["test_apples"]

    def setUp(self):
        self.purposes = DRFConstantChoiceField(Apple.purposes)
        self.colours = DRFConstantChoiceField(Apple.colours)

    def test_to_representation_int_backed(self):
        self.assertEqual(
            self.purposes.to_representation(Apple.purposes.cooking), u"cooking"
        )

    def test_to_representation_string_backed(self):
        self.assertEqual(
            self.colours.to_representation(Apple.colours.yellow), u"yellow"
        )

    def test_to_internal_value_int_backed(self):
        self.assertEqual(
            self.purposes.to_internal_value("eating"), Apple.purposes.eating
        )

    def test_to_internal_value_string_backed(self):
        self.assertEqual(self.colours.to_internal_value("red"), Apple.colours.red)

    def test_to_internal_value_not_a_valid_option_int_backed(self):
        with self.assertRaises(serializers.ValidationError) as e:
            self.colours.to_internal_value("purple")
        self.assertEqual(e.exception.detail, [u'"purple" is not a valid choice.'])

    def test_to_internal_value_not_a_valid_option_string_backed(self):
        with self.assertRaises(serializers.ValidationError) as e:
            self.purposes.to_internal_value("mincing")
        self.assertEqual(e.exception.detail, [u'"mincing" is not a valid choice.'])

    def test_in_serializer_output(self):
        instance = Apple.objects.all().first()
        serializer = AppleSerializer(instance=instance)
        self.assertEqual(
            serializer.data,
            {"name": "Granny Smith", "colour": "green", "purpose": "cooking"},
        )

    def test_in_serializer_input_bad(self):
        data = {"name": "Fuji", "colour": "blue", "purpose": "dicing"}
        serializer = AppleSerializer(data=data)
        self.assertFalse(serializer.is_valid())
        self.assertEqual(
            serializer.errors,
            {
                "colour": [u'"blue" is not a valid choice.'],
                "purpose": [u'"dicing" is not a valid choice.'],
            },
        )

    def test_in_serializer_input_good(self):
        data = {"name": "Fuji", "colour": "red", "purpose": "eating"}
        serializer = AppleSerializer(data=data)
        self.assertTrue(serializer.is_valid())
        instance = serializer.save()
        self.assertTrue(instance.colour.red)
        self.assertTrue(instance.purpose.eating)
        self.assertTrue(instance.purpose.culinary)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from django.db import models
from django.utils.translation import ugettext_lazy as _

from konst import Constant, ConstantGroup, Constants
from konst.models.fields import ConstantChoiceCharField, ConstantChoiceField


class Apple(models.Model):

    purposes = Constants(
        Constant(cooking=0, label=_("Cook me!")),
        Constant(eating=1, label=_("Eat me!")),
        Constant(juicing=2, label=_("Juice me!")),
        Constant(ornamental=3, label=_("Just look how pretty I am!")),
        ConstantGroup("culinary", ("cooking", "eating", "juicing")),
    )
    colours = Constants(
        Constant(red="FF0000", label=_("red")),
        Constant(green="00FF00", label=_("green")),
        Constant(yellow="FFFF80", label=_("yellow")),
        Constant(white="FFFFFF", label=_("white")),
    )

    name = models.CharField(max_length=30)
    purpose = ConstantChoiceField(constants=purposes)
    colour = ConstantChoiceCharField(constants=colours, max_length=30)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import json

from django.core.serializers.json import DjangoJSONEncoder

from konst import Constant


class ExtendedJSONEncoder(DjangoJSONEncoder):
    """Add support for serializing our class Constant."""

    def default(self, obj):
        if isinstance(obj, Constant):
            return obj.v
        else:
            return super(ExtendedJSONEncoder, self).default(obj)


def dumps(*args, **kwargs):
    kwargs["cls"] = kwargs.pop("cls", ExtendedJSONEncoder)
    return json.dumps(*args, **kwargs)


def loads(*args, **kwargs):
    return json.loads(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from django.utils import six
from django.utils.translation import ugettext_lazy as _

from rest_framework.fields import Field


class ConstantChoiceField(Field):
    """
    Extension of rest_framework.fields.ChoiceField.

    Allows exposure and receipt of string values from Constants.
    """

    default_error_messages = {"invalid_choice": _('"{input}" is not a valid choice.')}

    def __init__(self, constants, **kwargs):
        self.constants = constants
        self.allow_blank = kwargs.pop("allow_blank", False)
        super(ConstantChoiceField, self).__init__(**kwargs)

    def to_internal_value(self, data):
        """Given a string, convert to an integer value."""
        if data == "" and self.allow_blank:
            return ""
        try:
            return self.constants.by_id[six.text_type(data)]
        except KeyError:
            self.fail("invalid_choice", input=data)

    def to_representation(self, value):
        """Given a Constant instance, convert to the string representation."""
        if value in ("", None):
            return value
        # if the value is not mappable in our Constants then we've mucked up and
        # removed a Constant choice while leaving rows in our db with the removed
        # value. Rather than fail catastrophically, we indicate a current value
        # of null
        c = self.constants.by_value.get(value)
        if c:
            return c.id
        return None



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from django import forms


class ConstantChoiceField(forms.ChoiceField):
    def __init__(self, constants, *args, **kwargs):
        if "choices" in kwargs:
            raise ValueError(
                "`choices` is not supported by ConstantChoiceField. Use `constants` instead."
            )
        kwargs["choices"] = constants.choices
        super(ConstantChoiceField, self).__init__(*args, **kwargs)


__all__ = [ConstantChoiceField]



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from django.utils.deconstruct import deconstructible
from django.utils.encoding import python_2_unicode_compatible

# https://code.djangoproject.com/wiki/CookBookChoicesContantsClass
# modified to work with translation plus fields


@deconstructible
@python_2_unicode_compatible
class Constant(object):
    def __init__(self, label=None, **kwargs):
        self.constants = None
        assert len(kwargs) == 1
        for k, v in kwargs.items():
            self.id = k
            self.v = v
        self.label = label or self.id

    def __str__(self):
        return u"{}".format(self.v)

    def __repr__(self):
        return u"{} ({})".format(self.id, self.v)

    def __resolve_other(self, other):
        if isinstance(other, Constant):
            return other.v
        return other

    def __eq__(self, other):
        return self.v == self.__resolve_other(other)

    def __ne__(self, other):
        return self.v != self.__resolve_other(other)

    def __gt__(self, other):
        return self.v > self.__resolve_other(other)

    def __ge__(self, other):
        return self.v >= self.__resolve_other(other)

    def __lt__(self, other):
        return self.v < self.__resolve_other(other)

    def __le__(self, other):
        return self.v <= self.__resolve_other(other)

    def __hash__(self):
        if isinstance(self.v, str):
            return hash(self.v)
        return self.v

    def __getattr__(self, attr):
        """Allow dynamic lookup of equality and label."""
        if attr == u"constants":
            # https://nedbatchelder.com/blog/201010/surprising_getattr_recursion.html
            raise AttributeError(
                u"`constants` not set on Constant so cannot check equality"
            )
        if attr in self.constants.by_id:
            return self.v == self.constants.by_id[attr].v
        if attr in self.constants.groups:
            return self.id in self.constants.groups[attr].constant_ids
        else:
            raise AttributeError(
                u"'{}' object has no attribute '{}'".format(
                    self.__class__.__name__, attr
                )
            )

    def __len__(self):
        """Allow for underlying string values."""
        return len(self.v)

    def __bool__(self):
        """Evaluate all Constant instances to True."""
        return True

    __nonzero__ = __bool__


class ConstantGroup(object):
    def __init__(self, name, constant_ids):
        self.name = name
        self.constant_ids = set(constant_ids)


class Constants(object):
    def __init__(self, *args):
        self.constants = []
        self.groups = {}
        for a in args:
            if isinstance(a, Constant):
                self.constants.append(a)
                setattr(self, a.id, a)
                # add a reference to this class to the Constant instance
                setattr(a, "constants", self)
            elif isinstance(a, ConstantGroup):
                self.groups[a.name] = a
            else:
                raise ValueError(u"Received unexpected arg: {}".format(a))

        # setup useful precomputed lookups
        self.choices = [(k.v, k.label) for k in self.constants]
        self.by_value = dict([(k.v, k) for k in self.constants])
        self.by_id = dict([(k.id, k) for k in self.constants])

        # setup groups
        for name, group in self.groups.items():
            if hasattr(self, name):
                raise ValueError(
                    u"ConstantGroup with name `{}` clashes with existing attribute.".format(
                        name
                    )
                )
            setattr(self, name, set([self.by_id[i] for i in group.constant_ids]))



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from django.db.models import CharField, PositiveSmallIntegerField

from .. import Constant


class ConstantChoiceFieldMixin(object):
    def __init__(self, *args, **kwargs):
        self.constants = kwargs.pop("constants", None)
        if self.constants:
            kwargs["choices"] = self.constants.choices
        super(ConstantChoiceFieldMixin, self).__init__(*args, **kwargs)

    def to_python(self, value):
        # print "to_python: {}: {}".format(type(value), value)
        if value is None:
            return value
        if isinstance(value, Constant):
            return value
        return self.constants.by_value.get(int(value))

    def get_prep_value(self, value):
        # print "get_prep_value: {}: {}".format(type(value), value)
        if isinstance(value, Constant):
            return value.v
        return value

    def from_db_value(self, value, expression, connection, *args):
        # print "from_db_value: {}: {}".format(type(value), value)
        if value is None:
            return value
        # in some circumstances the field is loaded without constants
        if self.constants is None:
            return value
        return self.constants.by_value.get(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        value = super(ConstantChoiceFieldMixin, self).get_db_prep_value(
            value, connection, prepared
        )
        # print "get_db_prep_value: {}: {}".format(type(value), value)
        if value is None:
            return None
        elif isinstance(value, Constant):
            return value.v
        else:
            return value


class ConstantChoiceField(ConstantChoiceFieldMixin, PositiveSmallIntegerField):
    pass


class ConstantChoiceCharField(ConstantChoiceFieldMixin, CharField):
    def to_python(self, value):
        # print "to_python: {}: {}".format(type(value), value)
        if value is None:
            return value
        if isinstance(value, Constant):
            return value
        return self.constants.by_value.get(u"{}".format(value))


__all__ = [ConstantChoiceField, ConstantChoiceCharField]



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import os

import django

DEBUG = True
USE_TZ = True
TIME_ZONE = "UTC"
DATABASES = {
    "default": {
        "ENGINE": os.environ.get(
            "CONSTANTS_DATABASE_ENGINE", "django.db.backends.sqlite3"
        ),
        "HOST": os.environ.get("CONSTANTS_DATABASE_HOST", "127.0.0.1"),
        "NAME": os.environ.get("CONSTANTS_DATABASE_NAME", "konst"),
        "USER": os.environ.get("CONSTANTS_DATABASE_USER", ""),
    }
}
MIDDLEWARE = []  # from 2.0 onwards, only MIDDLEWARE is used

if django.VERSION < (1, 10):
    MIDDLEWARE_CLASSES = MIDDLEWARE
ROOT_URLCONF = "konst.tests.urls"
INSTALLED_APPS = ["konst", "konst.tests", "rest_framework"]
SITE_ID = 1

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": ["konst/tests/templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "debug": True,
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.template.context_processors.request",
            ],
        },
    }
]
SECRET_KEY = "konst-secret-key"
